unit ulogin;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Spin, ExtCtrls, Buttons, Menus, uTCPSockets, uSettings;

type

  { TfLogin }
  TfLogin = class(TForm)
    bConnect: TButton;
    bCancel: TButton;
    cbIP: TComboBox;
    edPass: TEdit;
    edUser: TEdit;
    ilProxy: TImageList;
    imProxy: TImage;
    lbIP: TLabel;
    lbPass: TLabel;
    lbPort: TLabel;
    lbUser: TLabel;
    pnBottom: TPanel;
    spPort: TSpinEdit;
    tmAutoConnect: TTimer;
    procedure bConnectClick(Sender: TObject);
    procedure cbIPChange(Sender: TObject);
    procedure cbIPCloseUp(Sender: TObject);
    procedure cbIPEnter(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure imProxyClick(Sender: TObject);
    procedure imProxyMouseEnter(Sender: TObject);
    procedure imProxyMouseLeave(Sender: TObject);
    procedure tmAutoConnectTimer(Sender: TObject);
  private
    FTCPClient: TTCPClient;
    FSettings: TSettings;
    FFirstTime: Boolean;
    FAutoConnectCnt: Integer;
    procedure EnableDisable(AEnable: Boolean);
    procedure LoadSettings;
    procedure SaveSettings;
  public
    constructor Create(TheOwner: TComponent; ATCPClient: TTCPClient; ASettings: TSettings); reintroduce;
  public
  end;

var
  fLogin: TfLogin;

implementation

uses uMain, uProxy, uCommon;

{$R *.lfm}

{ TfLogin }

constructor TfLogin.Create(TheOwner: TComponent; ATCPClient: TTCPClient;
  ASettings: TSettings);
begin
  inherited Create(TheOwner);
  FTCPClient := ATCPClient;
  FSettings := ASettings;
end;

procedure TfLogin.FormShow(Sender: TObject);
begin
  LoadSettings;
end;

procedure TfLogin.imProxyClick(Sender: TObject);
begin
  fProxy := TfProxy.Create(fLogin, FSettings.ActiveConnection);
  try
    fProxy.ShowModal;
  finally
    fProxy.Free;
  end;
end;

procedure TfLogin.imProxyMouseEnter(Sender: TObject);
begin
  ilProxy.GetBitmap(1, imProxy.Picture.Bitmap);
end;

procedure TfLogin.imProxyMouseLeave(Sender: TObject);
begin
  ilProxy.GetBitmap(0, imProxy.Picture.Bitmap);
end;

procedure TfLogin.EnableDisable(AEnable: Boolean);
begin
  cbIP.Enabled := AEnable;
  lbIP.Enabled := AEnable;
  spPort.Enabled := AEnable;
  lbPort.Enabled := AEnable;
  edUser.Enabled := AEnable;
  lbUser.Enabled := AEnable;
  edPass.Enabled := AEnable;
  lbPass.Enabled := AEnable;
  bConnect.Enabled := AEnable;
  bCancel.Enabled := AEnable;
end;

procedure TfLogin.cbIPCloseUp(Sender: TObject);
var
  Connection: TConnection;
begin
  if cbIP.ItemIndex <> -1 then
  begin
    Connection := TConnection(cbIP.Items.Objects[cbIP.ItemIndex]);
    if Connection = nil then
      Exit;
    if (Connection.FIP <> FSettings.ActiveConnection.FIP) or (FFirstTime) then
    begin
      spPort.Value := StrToIntDef(Connection.FPort, 40516);
      edUser.Text := Connection.FUser;
      edPass.Text := Connection.FPass;
      FSettings.CopyConnection(Connection, FSettings.ActiveConnection);
    end;
  end;
end;

procedure TfLogin.cbIPChange(Sender: TObject);
begin
  if cbIP.ItemIndex <> -1 then
    cbIPCloseUp(cbIP);
end;

procedure TfLogin.LoadSettings;
var
  Connection: TConnection;
  I: Integer;
begin
  with FSettings do
  begin
    Load;
    FFirstTime := True;
    cbIP.ItemIndex := -1;
    for I := 0 to ConnectionList.Count - 1 do
    begin
      Connection := TConnection(ConnectionList[I]);
      cbIP.Items.AddObject(Connection.FIP, Connection);
      if Connection.FIP = ActiveConnection.FIP then
        cbIP.ItemIndex := I;
    end;
    if (cbIP.ItemIndex = -1) and (cbIP.Items.Count > 0) then
      cbIP.ItemIndex := 0;

    cbIPCloseUp(cbIP);
    FFirstTime := False;

    if FSettings.AutoConnect then
    begin
      FAutoConnectCnt := 5;
      bConnect.Caption := 'Connect (' + IntToStr(FAutoConnectCnt) + ')';
      tmAutoConnect.Enabled := True;
    end;
  end;
end;

procedure TfLogin.SaveSettings;
var
  I: Integer;
  CanAdd: Boolean;
  Connection: TConnection;
begin
  with FSettings do
  begin
    ActiveConnection.FIP := cbIP.Text;
    ActiveConnection.FPort := IntToStr(spPort.Value);
    ActiveConnection.FUser := edUser.Text;
    ActiveConnection.FPass := edPass.Text;
    CanAdd := True;
    for I := 0 to ConnectionList.Count - 1  do
    begin
      if TConnection(ConnectionList[I]).FIP = ActiveConnection.FIP then
      begin
        CanAdd := False;
        Connection := TConnection(ConnectionList[I]);
        CopyConnection(ActiveConnection, Connection);
        Break;
      end;
    end;
    if CanAdd then
    begin
      Connection := TConnection.Create;
      CopyConnection(ActiveConnection, Connection);
      cbIP.ItemIndex := cbIP.Items.AddObject(Connection.FIP, Connection);
      ConnectionList.Add(Connection);
    end;
    Save;
  end;
end;

procedure TfLogin.bConnectClick(Sender: TObject);
var
  LogStr: String;
begin
  if tmAutoConnect.Enabled then
  begin
    bConnect.Caption := 'Connect';
    tmAutoConnect.Enabled := False;
  end;
  if Trim(cbIP.Text) = '' then
  begin
    cbIp.SetFocus;
    MessageDlgEx('Please enter a valid ip or host.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;
  if Trim(edUser.Text) = '' then
  begin
    edUser.SetFocus;
    MessageDlgEx('Please enter a username.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;

  SaveSettings;
  EnableDisable(False);
  Self.Invalidate;
  Application.ProcessMessages;
  FTCPClient.HideInternalMessages := FSettings.HideInternalMessages;
  FTCPClient.HideErrorMessages := FSettings.HideErrorMessages;
  FTCPClient.DownloadDirectory := FSettings.DownloadDirectory;
  FTCPClient.Connect(FSettings.ActiveConnection);
  EnableDisable(True);
  if FTCPClient.Connected then
    ModalResult := mrOk
  else
  begin
    LogStr := 'Cannot connect to server. ';
    fMain.sb.Tag := 2;
    fMain.sb.Panels[0].Text := LogStr;
    MessageDlgEx(LogStr + 'Error message: ' + sLineBreak + '"' + FTCPClient.ErrMsg + '(' + IntToStr(FTCPClient.ErrCode) + ')"', mtInformation, [mbOk], [], False, Self);
  end;
end;

procedure TfLogin.tmAutoConnectTimer(Sender: TObject);
begin
  Dec(FAutoConnectCnt);
  bConnect.Caption := 'Connect (' + IntToStr(FAutoConnectCnt) + ')';
  if FAutoConnectCnt = 0 then
  begin
    tmAutoConnect.Enabled := False;
    bConnect.Caption := 'Connect';
    bConnectClick(bConnect);
  end;
end;

procedure TfLogin.cbIPEnter(Sender: TObject);
begin
  if (FSettings.AutoConnect) and (tmAutoConnect.Enabled) then
  begin
    tmAutoConnect.Enabled := False;
    bConnect.Caption := 'Connect';
  end;
end;


end.

