unit uMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ComCtrls,
  ExtCtrls, LazFileUtils, laz.VirtualTrees, Menus, Buttons, ActnList,
  LCLType, LCLIntf, uTCPSockets, uSettings, uCommon;

type

  { TfMain }
  TfMain = class(TForm)
    acQuit: TAction;
    acAbout: TAction;
    acStop: TAction;
    acStart: TAction;
    alMain: TActionList;
    acOptions: TAction;
    imSB: TImageList;
    lbLog: TListBox;
    MenuItem1: TMenuItem;
    miRemoveZombies: TMenuItem;
    miSepDisconnect: TMenuItem;
    miBroadcastScreen: TMenuItem;
    miBroadCastFile: TMenuItem;
    miBroadcastMessage: TMenuItem;
    miBroadcast: TMenuItem;
    miDisconnect: TMenuItem;
    miClear: TMenuItem;
    miHide: TMenuItem;
    miHideErrorMessages: TMenuItem;
    miHideInternalMessages: TMenuItem;
    miOpen: TMenuItem;
    miSave: TMenuItem;
    mm: TMainMenu;
    miAbout: TMenuItem;
    miHelp: TMenuItem;
    miStop: TMenuItem;
    miStart: TMenuItem;
    miServer: TMenuItem;
    miFile: TMenuItem;
    miTools: TMenuItem;
    miOptions: TMenuItem;
    miQuit: TMenuItem;
    OD: TOpenDialog;
    pmLog: TPopupMenu;
    pnLog: TPanel;
    pnMain: TPanel;
    pnThreads: TPanel;
    pnInfo: TPanel;
    pmClients: TPopupMenu;
    sb: TStatusBar;
    SD: TSaveDialog;
    sp: TSplitter;
    tmWait: TTimer;
    VSTClients: TLazVirtualStringTree;
    procedure acQuitExecute(Sender: TObject);
    procedure acOptionsExecute(Sender: TObject);
    procedure acAboutExecute(Sender: TObject);
    procedure acStartExecute(Sender: TObject);
    procedure acStopExecute(Sender: TObject);
    procedure edBindIPKeyPress(Sender: TObject; var Key: char);
    procedure FormCloseQuery(Sender: TObject; var {%H-}CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure miBroadcastMessageClick(Sender: TObject);
    procedure miBroadCastFileClick(Sender: TObject);
    procedure miClearClick(Sender: TObject);
    procedure miDisconnectClick(Sender: TObject);
    procedure miHideErrorMessagesClick(Sender: TObject);
    procedure miHideInternalMessagesClick(Sender: TObject);
    procedure miRemoveZombiesClick(Sender: TObject);
    procedure miSaveClick(Sender: TObject);
    procedure miOpenClick(Sender: TObject);
    procedure miBroadcastScreenClick(Sender: TObject);
    procedure pmLogPopup(Sender: TObject);
    procedure sbDrawPanel(StatusBar: TStatusBar; Panel: TStatusPanel;
      const Rect: TRect);
    procedure tmWaitTimer(Sender: TObject);
    procedure VSTClientsAfterCellPaint(Sender: TBaseVirtualTree;
      TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
      const CellRect: TRect);
    procedure VSTClientsCompareNodes(Sender: TBaseVirtualTree; Node1,
      Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
    procedure VSTClientsDblClick(Sender: TObject);
    procedure VSTClientsFreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure VSTClientsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType; var CellText: String);
    procedure VSTClientsHeaderClick(Sender: TVTHeader; HitInfo: TVTHeaderHitInfo
      );
    procedure VSTClientsMouseUp(Sender: TObject; Button: TMouseButton;
      {%H-}Shift: TShiftState; X, Y: Integer);
  private
    FSortCol: Integer;
    FSortDir: TSortDirection;
    FTCPServer: TTCPServer;
    FSettings: TSettings;
    FScreenShot: TScreenShot;
    FScreenShotUniqueIDs: array of Integer;
    FBroadcastingScreen: Boolean;
    function AddClientNode(AConnection: TConnection): PVirtualNode;
    function FindClientNode(AUniqueID: Integer): PVirtualNode;
    function AddClientPage(ANode: PVirtualNode; AConnection: TConnection; const AShow: Boolean = True): TTabSheet;
    function FindClientPage(ANode: PVirtualNode; const AShow: Boolean = False): TTabSheet;
    function FindClientFrame(AConnection: TConnection): TFrame;
    procedure Cleanup;
    procedure AddLog(ALogMsg: String);
    procedure BeginBroadcastScreen;
    procedure EndBroadcastScreen;
    procedure DoOnConnect(Sender: TObject; AConnection: TConnection);
    procedure DoOnDisconnect(Sender: TObject; AConnection: TConnection; {%H-}AByUser: Boolean);
    procedure DoOnError(Sender: TObject; AErrMsg: string; AErrCode: Integer; {%H-}AConnection: TConnection);
    procedure DoOnInternalMessage(Sender: TObject; AMsg: String);
    procedure DoOnProgress(Sender: TObject; AStreamingType: TStreamingType; ACnt, ATotCnt: Int64; ASpeed, ARemaining: LongInt; AConnection: TConnection);
    procedure DoOnRecvMessage(Sender: TObject; AMsg: String; AParams: TStringArray; AConnection: TConnection);
    procedure DoOnRecvStream(Sender: TObject; AMsg: String; AParams: TStringArray; AMS: TMemoryStream; AConnection: TConnection);
    procedure DoOnRecvFile(Sender: TObject; AMsg: String; AParams: TStringArray; AFileName: String; AConnection: TConnection);
    procedure DoOnScreenShot(Sender: TObject; AMS: TMemoryStream);
  public
  end;

const
  Version = '1.0';
  SynTrunkRev = 209;

var
  fMain: TfMain;

implementation

uses uOptions, uStart, uClients, uAbout, uTCPFunctions, uClient, uBroadcastMessage;

{$R *.lfm}

{ TfMain }

type
  PData = ^TData;
  TData = record
    FIP, FHost, FUser: String;
    FUniqueID: Integer;
    FProgress: Integer;
    FProgressStr: String;
    FStreamingType: TStreamingType;
    FSpeed: String;
    FRemaining: String;
    FSize: String;
    FTotalSize: String;
    FConnection: TConnection;
    FTSClient: TTabSheet;
  end;

procedure TfMain.FormCreate(Sender: TObject);
begin
  FSettings := TSettings.Create;
  FSettings.FileName := AppendPathDelim(ExtractFilePath(Application.Exename)) + 'conf.xml';

  FTCPServer := TTCPServer.Create;
  FTCPServer.OnConnect := @DoOnConnect;
  FTCPServer.OnDisconnect := @DoOnDisconnect;
  FTCPServer.OnError := @DoOnError;
  FTCPServer.OnInternalMessage := @DoOnInternalMessage;
  FTCPServer.OnProgress := @DoOnProgress;
  FTCPServer.OnRecvMessage := @DoOnRecvMessage;
  FTCPServer.OnRecvStream := @DoOnRecvStream;
  FTCPServer.OnRecvFile := @DoOnRecvFile;
  VSTClients.NodeDataSize := SizeOf(TData);

  sb.Tag := 0;
  sb.Panels[0].Text := 'Idle. Waiting for commands.';
end;

procedure TfMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if FBroadCastingScreen then
    EndBroadcastScreen;
  Cleanup;
end;

procedure TfMain.FormDestroy(Sender: TObject);
begin
  FTCPServer.Free;
  FSettings.Free;
end;

procedure TfMain.FormShow(Sender: TObject);
begin
  tmWait.Enabled := True;
end;

procedure TfMain.tmWaitTimer(Sender: TObject);
begin
  tmWait.Enabled := False;
  acStartExecute(acStart);
end;

function TfMain.AddClientNode(AConnection: TConnection): PVirtualNode;
var
  Data: PData;
begin
  Result := VSTClients.AddChild(nil);
  if Result = nil then
    Exit;
  Data := VSTClients.GetNodeData(Result);
  Data^.FIP := AConnection.FExtendedInfo.FRemoteIP;
  Data^.FHost := AConnection.FExtendedInfo.FRemoteHost;
  Data^.FUser := AConnection.FUser;
  Data^.FUniqueID := AConnection.FUniqueID;
  Data^.FProgress := -1;
  Data^.FSpeed := '';
  Data^.FSize := '';
  Data^.FTotalSize := '';
  Data^.FTSClient := AddClientPage(Result, AConnection, False);
end;

function TfMain.FindClientNode(AUniqueID: Integer): PVirtualNode;
var
  Node: PVirtualNode;
  Data: PData;
begin
  Result := nil;
  Node := VSTClients.GetFirst;
  while Assigned(Node) do
  begin
    Data := VSTClients.GetNodeData(Node);
    if (Data^.FUniqueID = AUniqueID) then
    begin
      Result := Node;
      Break;
    end;
    Node := VSTClients.GetNext(Node);
  end;
end;

function TfMain.FindClientPage(ANode: PVirtualNode;
  const AShow: Boolean = False): TTabSheet;
var
  Data: PData;
begin
  Result := nil;
  if ANode <> nil then
  begin
    Data := VSTClients.GetNodeData(ANode);
    Result := Data^.FTSClient;
    if (Result <> nil) then
    begin
      Result.TabVisible := True;
      if AShow then
      begin
        fClients.pgClients.ActivePage := Data^.FTSClient;
        if fClients.Visible then
          fClients.SetFocus
        else
          fClients.ShowModal;
      end;
    end;
  end;
end;

function TfMain.FindClientFrame(AConnection: TConnection): TFrame;
var
  Node: PVirtualNode;
  Data: PData;
begin
  Result := nil;
  Node := PVirtualNode(AConnection.FData);
  if Node <> nil then
  begin
    Data := VSTClients.GetNodeData(Node);
    if Data^.FTSClient <> nil then
      Result := TFrame(Data^.FTSClient.FindChildControl('frClient'));
  end;
end;

function TfMain.AddClientPage(ANode: PVirtualNode; AConnection: TConnection; const AShow: Boolean): TTabSheet;
var
  Data: PData;
  ActivePageIndex: Integer;
begin
  Result := nil;
  if ANode <> nil then
  begin
    Data := VSTClients.GetNodeData(ANode);
    Result := fClients.FindClientPage(Data^.FUniqueID);
    if Result = nil then
    begin
      ActivePageIndex := fClients.pgClients.ActivePageIndex;
      Result := fClients.AddClientPage(FTCPServer, FSettings, AConnection);
      if fClients.Visible then
      begin
        if ActivePageIndex <> - 1 then
          fClients.pgClients.ActivePageIndex := ActivePageIndex;
      end
      else
        fClients.pgClients.ActivePage := Result;
      if AShow then
      begin
        if fClients.Visible then
          fClients.SetFocus
        else
          fClients.ShowModal;
      end;
    end;
  end;
end;

procedure TfMain.miClearClick(Sender: TObject);
begin
  lbLog.Clear;
end;

procedure TfMain.miHideInternalMessagesClick(Sender: TObject);
begin
  miHideInternalMessages.Checked := not miHideInternalMessages.Checked;
  FSettings.HideInternalMessages := miHideInternalMessages.Checked;
  FTCPServer.HideInternalMessages := miHideInternalMessages.Checked;
  FSettings.Save;
end;

procedure TfMain.miHideErrorMessagesClick(Sender: TObject);
begin
  miHideErrorMessages.Checked := not miHideErrorMessages.Checked;
  FSettings.HideErrorMessages := miHideErrorMessages.Checked;
  FTCPServer.HideErrorMessages := miHideErrorMessages.Checked;
  FSettings.Save;
end;

procedure TfMain.pmLogPopup(Sender: TObject);
begin
  FSettings.Load;
  miHideInternalMessages.Checked := FSettings.HideInternalMessages;
  miHideErrorMessages.Checked := FSettings.HideErrorMessages;
end;

procedure TfMain.miSaveClick(Sender: TObject);
begin
  SD.FileName := 'Log';
  if SD.Execute then
    lbLog.Items.SaveToFile(SD.FileName);
end;

procedure TfMain.miOpenClick(Sender: TObject);
var
  Node: PVirtualNode;
begin
  if VSTClients.SelectedCount = 0 then
  begin
    MessageDlgEx('Please select a client first.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;
  Node := VSTClients.GetFirstSelected;
  while Assigned(Node) do
  begin
    if FindClientPage(Node, True) = nil then
      ShowMessage('Cannot locate client');//should never happen
    Node := VSTClients.GetNextSelected(Node);
  end;
end;

procedure TfMain.miBroadcastMessageClick(Sender: TObject);
var
  Node: PVirtualNode;
  Data: PData;
begin
  if VSTClients.SelectedCount = 0 then
  begin
    MessageDlgEx('Please select a client first.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;

  fBroadcastMessage := TfBroadcastMessage.Create(nil);
  try
    fBroadcastMessage.ShowModal;
    if fBroadcastMessage.ModalResult = mrOk then
    begin
      Node := VSTClients.GetFirstSelected;
      while Assigned(Node) do
      begin
        Data := VSTClients.GetNodeData(Node);
        FTCPServer.SendMessage('BROADCASTMESSAGE', [fBroadcastMessage.edMessage.Text, fBroadcastMessage.edParams.Text], Data^.FUniqueID);
        Node := VSTClients.GetNextSelected(Node);
      end;
    end;
  finally
    fBroadcastMessage.Free;
  end;
end;

procedure TfMain.miBroadCastFileClick(Sender: TObject);
var
  Node: PVirtualNode;
  Data: PData;
begin
  if VSTClients.SelectedCount = 0 then
  begin
    MessageDlgEx('Please select a client first.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;

  OD.InitialDir := FSettings.DownloadDirectory;
  if OD.Execute then
  begin
    Node := VSTClients.GetFirstSelected;
    while Assigned(Node) do
    begin
      Data := VSTClients.GetNodeData(Node);
      FTCPServer.SendFile('BROADCASTFILE', [], OD.FileName, Data^.FUniqueID);
      Node := VSTClients.GetNextSelected(Node);
    end;
  end;
end;

procedure TfMain.DoOnScreenShot(Sender: TObject; AMS: TMemoryStream);
var
  I: Integer;
  MS: TMemoryStream;
  TmpArray: array of Integer;
begin
  if Length(FScreenShotUniqueIDs) = 0 then
    Exit;
  TmpArray := nil;
  for I := Low(FScreenShotUniqueIDs) to High(FScreenShotUniqueIDs) do
  begin
    AMS.Position := 0;
    MS := TMemoryStream.Create;
    MS.CopyFrom(AMS, AMS.Size);
    MS.Position := 0;
    if FTCPServer.SendStream('BROADCASTSCREEN', [], MS, FScreenShotUniqueIDs[I]) then
    begin
      SetLength(TmpArray, Length(TmpArray) + 1);
      TmpArray[Length(TmpArray) - 1] := FScreenShotUniqueIDs[I];
    end
    else
      MS.Free;
  end;
  FScreenShotUniqueIDs := nil;
  for I := Low(TmpArray) to High(TmpArray) do
  begin
    SetLength(FScreenShotUniqueIDs, Length(FScreenShotUniqueIDs) + 1);
    FScreenShotUniqueIDs[Length(FScreenShotUniqueIDs) - 1] := TmpArray[I];
  end;
end;

procedure TfMain.BeginBroadcastScreen;
var
  Node: PVirtualNode;
  Data: PData;
  Cnt: Integer;
begin
  if VSTClients.SelectedCount = 0 then
  begin
    MessageDlgEx('Please select a client first.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;
  SetLength(FScreenShotUniqueIDs, VSTClients.SelectedCount);
  Cnt := 0;
  Node := VSTClients.GetFirstSelected;
  while Assigned(Node) do
  begin
    Data := VSTClients.GetNodeData(Node);
    FScreenShotUniqueIDs[Cnt] := Data^.FUniqueID;
    Inc(Cnt);
    Node := VSTClients.GetNextSelected(Node);
  end;
  if Cnt > 0 then
  begin
    FBroadcastingScreen := True;
    miBroadcastScreen.Caption := 'Screen - stop';
    miOpen.Enabled := False;
    miBroadcastMessage.Enabled := False;
    miBroadcastFile.Enabled := False;
    miDisconnect.Enabled := False;
    miRemoveZombies.Enabled :=  False;
    FScreenShot := TScreenShot.Create(pf24bit, 75);
    FScreenShot.OnScreenShot := @DoOnScreenShot;
    FScreenShot.Start;
  end;
end;

procedure TfMain.EndBroadcastScreen;
begin
  FBroadcastingScreen := False;
  miBroadcastScreen.Caption := 'Screen';
  miOpen.Enabled := True;
  miBroadcastMessage.Enabled := True;
  miBroadcastFile.Enabled := True;
  miDisconnect.Enabled := True;
  miRemoveZombies.Enabled := True;
  if FScreenShot <> nil then
  begin
    FScreenShot.OnScreenShot := nil;
    FScreenShot.Terminate;
    FScreenShot.WaitFor;
    FScreenShot.Free;
    FScreenShot := nil;
  end;
end;

procedure TfMain.miBroadcastScreenClick(Sender: TObject);
begin
  if not FBroadcastingScreen then
    BeginBroadcastScreen
  else
    EndBroadcastScreen;
end;

procedure TfMain.miDisconnectClick(Sender: TObject);
var
  Node: PVirtualNode;
  Data: PData;
  Cnt, I: Integer;
  UniqueIDs: array of Integer;
begin
  if VSTClients.SelectedCount = 0 then
  begin
    MessageDlgEx('Please select a client first.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;
  UniqueIDs := nil;
  SetLength(UniqueIDs, VSTClients.SelectedCount);
  Cnt := 0;
  Node := VSTClients.GetFirstSelected;
  while Assigned(Node) do
  begin
    Data := VSTClients.GetNodeData(Node);
    UniqueIDs[Cnt] := Data^.FUniqueID;
    Inc(Cnt);
    Node := VSTClients.GetNextSelected(Node);
  end;
  if Cnt > 0 then
    for I := Low(UniqueIDs) to High(UniqueIDs) do
      FTCPServer.DisconnectClient(UniqueIDs[I]);
end;

procedure TfMain.miRemoveZombiesClick(Sender: TObject);
var
  Node: PVirtualNode;
  Data: PData;
  I: Integer;
  NodeArray: array of PVirtualNode;
begin
  NodeArray := nil;
  Node := VSTClients.GetFirst;
  while Assigned(Node) do
  begin
    Data := VSTClients.GetNodeData(Node);
    if not FTCPServer.IsClientConnected(Data^.FUniqueID) then
    begin
      SetLength(NodeArray, Length(NodeArray) + 1);
      NodeArray[Length(NodeArray) - 1] := Node;
    end;
    Node := VSTClients.GetNext(Node);
  end;
  for I := Low(NodeArray) to High(NodeArray) do
    VSTClients.DeleteNode(NodeArray[I]);
  if Length(NodeArray) = 0 then
    MessageDlgEx('No zombies found.', mtInformation, [mbOK], [], False, Self)
  else
    MessageDlgEx(IntToStr(Length(NodeArray)) + ' zombies removed.', mtInformation, [mbOK], [], False, Self);
end;

procedure TfMain.sbDrawPanel(StatusBar: TStatusBar; Panel: TStatusPanel;
  const Rect: TRect);
var
  H, T, L: Integer;
begin
  with StatusBar.Canvas do
  begin
    H := StatusBar.Canvas.GetTextHeight('getheight');
    T := Rect.Top + (Rect.Bottom - Rect.Top - H) div 2;
    L := Rect.Left + 5;
    case Panel.Index of
      0: begin
           case sb.Tag of
             0: imSB.Draw(StatusBar.Canvas, L, T, 0);
             1: imSB.Draw(StatusBar.Canvas, L, T, 1);
             2: imSB.Draw(StatusBar.Canvas, L, T, 2);
           end;
           TextRect(Rect, L + ImSb.Width + 5, T, Panel.Text);
        end;
    end;
  end;
end;

procedure TfMain.edBindIPKeyPress(Sender: TObject; var Key: char);
begin
  if not (Key in ['0'..'9', '.', #8, #9]) then
    Key := #0;
end;

procedure TfMain.Cleanup;
var
  TmpCap: String;
begin
  TmpCap := Caption;
  if VSTClients.RootNodeCount > 0 then
    Caption := 'Closing connections. Plese wait...'
  else
    Caption := 'Stopping server. Please wait';
  Application.ProcessMessages;
  FTCPServer.Stop;
  Caption := TmpCap;
  if FScreenShot <> nil then
  begin
    FScreenShot.Terminate;
    FScreenShot.WaitFor;
    FScreenShot := nil;
  end;
end;

procedure TfMain.AddLog(ALogMsg: String);
begin
  lbLog.Items.Add(FormatDateTime('hh:mm:ss', now) + '  ' + ALogMsg);
  lbLog.TopIndex := lbLog.Items.Count - 1;
end;

procedure TfMain.DoOnConnect(Sender: TObject; AConnection: TConnection);
var
  Node: PVirtualNode;
begin
  Node := AddClientNode(AConnection);
  AConnection.FData := Node;
  VSTClients.SortTree(FSortCol, FSortDir);
  AddLog('Connection accepted from: ' + AConnection.FExtendedInfo.FRemoteIP + '(' + AConnection.FExtendedInfo.FRemoteHost + '). ' + '"' + AConnection.FUser + '" has successfully loged in.');
end;

procedure TfMain.DoOnDisconnect(Sender: TObject; AConnection: TConnection;
  AByUser: Boolean);
var
  Node: PVirtualNode;
  tsClient: TTabSheet;
begin
  if (FBroadcastingScreen) and (Length(FScreenShotUniqueIDs) = 1) and (FScreenShotUniqueIDs[0] = AConnection.FUniqueID) then
    EndBroadcastScreen;

  Node := FindClientNode(AConnection.FUniqueID);
  if Node <> nil then
  begin
    tsClient := fClients.FindClientPage(AConnection.FUniqueID);
    if tsClient <> nil then
    begin
      tsClient.Free;
      tsClient := nil;
    end;
    VSTClients.DeleteNode(Node);
  end;
  if FTCPServer.Active then
    AddLog(AConnection.FExtendedInfo.FRemoteIP + '(' + AConnection.FExtendedInfo.FRemoteHost + ') has disconnected.');
  Application.ProcessMessages;
end;

procedure TfMain.DoOnError(Sender: TObject; AErrMsg: string; AErrCode: Integer;
  AConnection: TConnection);
begin
  if FSettings.HideErrorMessages then
    Exit;
  if AConnection.FUniqueID = 0 then
    AddLog('Sever error. ' + AErrMsg + '(' + IntToStr(AErrCode) + ')')
  else
    AddLog('Error for ' + AConnection.FExtendedInfo.FRemoteIP + ' : ' + AConnection.FUser + '. ' + AErrMsg + '(' + IntToStr(AErrCode) + ')');
end;

procedure TfMain.DoOnInternalMessage(Sender: TObject; AMsg: String);
begin
  if FSettings.HideInternalMessages then
    Exit;
  AddLog(AMsg);
end;

procedure TfMain.DoOnProgress(Sender: TObject; AStreamingType: TStreamingType;
  ACnt, ATotCnt: Int64; ASpeed, ARemaining: LongInt; AConnection: TConnection);
var
  SpeedStr: String;
  Node: PVirtualNode;
  Data: PData;
begin
  Node := PVirtualNode(AConnection.FData);
  if Node <> nil then
  begin
    Data := VSTClients.GetNodeData(Node);

    Data^.FProgress := Round(ACnt*100/ATotCnt);
    case AStreamingType of
      stDownload: Data^.FProgressStr := 'Downloading ' + IntToStr(Data^.FProgress) + '%';
      stUpload: Data^.FProgressStr := 'Uploading ' + IntToStr(Data^.FProgress) + '%';
    end;
    if Data^.FProgress >= 100 then
      Data^.FProgress := -1;
    Data^.FStreamingType := AStreamingType;

    if ACnt < ATotCnt then
      SpeedStr := FormatSpeed(ASpeed)
    else
     SpeedStr := '';
    Data^.FSpeed := SpeedStr;

    if ACnt < ATotCnt then
       Data^.FRemaining := SecToHourAndMin(ARemaining)
     else
       Data^.FRemaining := '';

    if ACnt < ATotCnt then
      Data^.FSize := FormatSize(ACnt)
    else
      Data^.FSize := '';

    if ACnt < ATotCnt then
      Data^.FTotalSize := FormatSize(ATotCnt)
    else
      Data^.FTotalSize := '';

    VSTClients.ReinitNode(Node, False);
    VSTClients.RepaintNode(Node);
  end;
end;

procedure TfMain.DoOnRecvMessage(Sender: TObject; AMsg: String; AParams: TStringArray;
  AConnection: TConnection);
type
  TAddLog = procedure(ALogMsg: String) of Object;
var
  frClient: TfrClient;
  AddLogCallback: TAddLog;
begin
  frClient := TfrClient(FindClientFrame(AConnection));
  if frClient <> nil then
  begin
    AddLogCallback := @AddLog;
    frClient.RecvMessage(Sender, AMsg, AParams, AConnection, AddLogCallback);
    if not fClients.Visible then
      fClients.ShowModal;
  end;
end;

procedure TfMain.DoOnRecvStream(Sender: TObject; AMsg: String;
  AParams: TStringArray; AMS: TMemoryStream; AConnection: TConnection);
type
  TAddLog = procedure(ALogMsg: String) of Object;
var
  frClient: TfrClient;
  AddLogCallback: TAddLog;
begin
  frClient := TfrClient(FindClientFrame(AConnection));
  if frClient <> nil then
  begin
    AddLogCallback := @AddLog;
    frClient.RecvStream(Sender, AMsg, AParams, AMS, AConnection, AddLogCallback);
    if not fClients.Visible then
      fClients.ShowModal;
  end;
end;

procedure TfMain.DoOnRecvFile(Sender: TObject; AMsg: String;
  AParams: TStringArray; AFileName: String; AConnection: TConnection);
type
  TAddLog = procedure(ALogMsg: String) of Object;
var
  frClient: TfrClient;
  AddLogCallback: TAddLog;
begin
  frClient := TfrClient(FindClientFrame(AConnection));
  if frClient <> nil then
  begin
    AddLogCallback := @AddLog;
    frClient.RecvFile(Sender, AMsg, AParams, AFileName, AConnection, AddLogCallback);
    if not fClients.Visible then
      fClients.ShowModal;
  end;
end;

procedure TfMain.acQuitExecute(Sender: TObject);
begin
  Close;
end;

procedure TfMain.acOptionsExecute(Sender: TObject);
begin
  fOptions := TfOptions.Create(nil, FSettings);
  try
    fOptions.ShowModal;
    FTCPServer.HideInternalMessages := FSettings.HideInternalMessages;
    FTCPServer.HideErrorMessages := FSettings.HideErrorMessages;
    FTCPServer.DownloadDirectory := FSettings.DownloadDirectory;
  finally
    fOptions.Free;
  end;
end;

procedure TfMain.acStartExecute(Sender: TObject);
var
  LogStr: String;
begin
  fStart := TfStart.Create(nil, FTCPServer, FSettings);
  try
    pnMain.Visible := False;
    fStart.ShowModal;
    if fStart.ModalResult = mrOk then
    begin
      acStart.Enabled := False;
      acStop.Enabled := True;
      if Trim(FSettings.BindConnection.FPass) <> '' then
        LogStr := 'Server successfully started, listening on port ' + FSettings.BindConnection.FPort + '. Password protection is on.'
      else
        LogStr := 'Server successfully started, listening on port ' + FSettings.BindConnection.FPort + '. Password protection is off.';
      AddLog(LogStr);
      sb.Tag := 1;
      sb.Panels[0].Text := LogStr;
      Caption := 'Server - active';
    end;
    pnMain.Visible := True;
  finally
    fStart.Free;
  end;
end;

procedure TfMain.acStopExecute(Sender: TObject);
var
  LogStr: String;
begin
  Cleanup;
  acStart.Enabled := True;
  acStop.Enabled := False;
  LogStr := 'Server stopped.';
  AddLog(LogStr);
  sb.Tag := 2;
  sb.Panels[0].Text := LogStr;
  Caption := 'Server - not active';
end;

procedure TfMain.acAboutExecute(Sender: TObject);
begin
  fAbout := TfAbout.Create(nil);
  try
    fAbout.ShowModal;
  finally
    fAbout.Free;
  end;
end;

procedure TfMain.VSTClientsAfterCellPaint(Sender: TBaseVirtualTree;
  TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
  const CellRect: TRect);
const
  BackgroundStartColor = $00FCFCFC;
  BackgroundStopColor = $00C8C8C8;
  DownloadColor = $024A60D;
  UploadColor = $001C9EFF;
var
  Data: PData;
  W: Integer;
  Str: String;
  Rect, DrawRect: TRect;
  TextColor: TColor;
begin
  if Column = 4 then
  begin
    Data := VSTClients.GetNodeData(Node);
    if Data^.FProgress = -1 then
      Exit;
    Rect.Create(CellRect);
    InflateRect(Rect, -2, -2);
    DrawRect.Create(Rect);
    with TargetCanvas do
    begin
      Brush.Color := clGray;
      FillRect(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
      if (Node = Sender.FocusedNode) then
        Brush.Color := $00D7D7D7
      else
        Brush.Color := clWhite;
      FillRect(Rect.Left + 1, Rect.Top + 1, Rect.Right - 1, Rect.Bottom - 1);

      SetRect(DrawRect, Rect.Left + 2, Rect.Top + 2, Rect.Right - 2, Rect.Bottom - 2);
      if (Node = Sender.FocusedNode) then
        GradientFill(DrawRect, VSTClients.Colors.SelectionRectangleBorderColor, VSTClients.Colors.SelectionRectangleBorderColor, gdHorizontal)
      else
        GradientFill(DrawRect, BackgroundStartColor, BackgroundStopColor, gdHorizontal);

      if (Data^.FProgress > 0) then
      begin
        if (Data^.FProgress > 100) then
          Data^.FProgress := 100;
        W := Round(Rect.Width*Data^.FProgress/100);
        SetRect(DrawRect, Rect.Left + 2, Rect.Top + 2, Rect.Left + W - 2, Rect.Bottom - 2);
        case Data^.FStreamingType of
          stDownload: Brush.Color := DownloadColor;
          stUpload: Brush.Color := UploadColor;
        end;
        FillRect(DrawRect);
      end;

      SetRect(DrawRect, Rect.Left + 2, Rect.Top + 2, Rect.Right - 2, Rect.Bottom - 2);
      SetBkMode(Handle, TRANSPARENT);
      Str := Data^.FProgressStr;
      TextColor := Font.Color;
      if Data^.FProgress < 100 then
        Font.Color := clBlack
      else
        Font.Color := TextColor;
      DrawText(Handle, PChar(Str), -1, Rect, DT_SINGLELINE or DT_CENTER or DT_VCENTER or DT_END_ELLIPSIS or DT_NOPREFIX);
    end;
  end;
end;

procedure TfMain.VSTClientsCompareNodes(Sender: TBaseVirtualTree; Node1,
  Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
var
  Data1: PData;
  Data2: PData;
begin
  Data1 := Sender.GetNodeData(Node1);
  Data2 := Sender.GetNodeData(Node2);
  case Column of
    0: Result := CompareText(Data1^.FIP, Data2^.FIP);
    1: Result := CompareText(Data1^.FHost, Data2^.FHost);
    2: Result := CompareText(Data1^.FUser, Data2^.FUser);
  end;
end;

procedure TfMain.VSTClientsDblClick(Sender: TObject);
begin
  if not FBroadcastingScreen then
    miOpenClick(miOpen);
end;

procedure TfMain.VSTClientsFreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
var
  Data: PData;
begin
  Data := VSTClients.GetNodeData(Node);
  if Assigned(Data^.FTSClient) then
    Data^.FTSClient := nil;
  Finalize(Data^);
end;

procedure TfMain.VSTClientsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex; TextType: TVSTTextType; var CellText: String);
var
  Data: PData;
begin
  if TextType = ttStatic then
    Exit;
  Data := VSTClients.GetNodeData(Node);
  case Column of
    0: CellText := Data^.FIP;
    1: CellText := Data^.FHost;
    2: CellText := Data^.FUser;
    3: CellText := IntToStr(Data^.FUniqueID);
    4: CellText := '';
    5: CellText := Data^.FSpeed;
    6: CellText := Data^.FRemaining;
    7: CellText := Data^.FSize;
    8: CellText := Data^.FTotalSize;
  end;
  Application.Exename
end;

procedure TfMain.VSTClientsHeaderClick(Sender: TVTHeader;
  HitInfo: TVTHeaderHitInfo);
begin
  if (HitInfo.Column > 2) then
    Exit;
  if HitInfo.Button = mbLeft then
  begin
    with Sender, Treeview do
    begin
      if (SortColumn = NoColumn) or (SortColumn <> HitInfo.Column) then
      begin
        SortColumn    := HitInfo.Column;
        SortDirection := sdAscending;
      end
      else
      begin
        if SortDirection = sdAscending then
          SortDirection := sdDescending
        else
          SortDirection := sdAscending;
        FSortDir := SortDirection;
      end;
      SortTree(SortColumn, SortDirection, False);
      FSortCol := Sender.SortColumn;
    end;
  end;
end;

procedure TfMain.VSTClientsMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  Node: PVirtualNode;
  P: TPoint;
begin
  if Button = mbRight then
  begin
    Node := VSTClients.GetNodeAt(X, Y);
    if (Node <> nil) and VSTClients.Selected[Node] then
    begin
      P.X := X; P.Y := Y;
      P := ClientToScreen(P);
      pmClients.PopUp(P.X, P.Y);
    end;
  end;
end;

end.

