unit uStart;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Spin, ExtCtrls, Buttons, Menus, uTCPSockets, uSettings;

type

  { TfStart }
  TfStart = class(TForm)
    bStart: TButton;
    bCancel: TButton;
    edBindIP: TEdit;
    edPass: TEdit;
    lbBindIP: TLabel;
    lbPass: TLabel;
    lbPort: TLabel;
    pnBottom: TPanel;
    spPort: TSpinEdit;
    tmAutoStart: TTimer;
    procedure bStartClick(Sender: TObject);
    procedure edBindIPEnter(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure tmAutoStartTimer(Sender: TObject);
  private
    FTCPServer: TTCPServer;
    FSettings: TSettings;
    FAutoStartCnt: Integer;
    procedure EnableDisable(AEnable: Boolean);
    procedure LoadSettings;
    procedure SaveSettings;
  public
    constructor Create(TheOwner: TComponent; ATCPServer: TTCPServer; ASettings: TSettings); reintroduce;
  public
  end;

var
  fStart: TfStart;

implementation

uses uMain, uCommon;

{$R *.lfm}

{ TfStart }

constructor TfStart.Create(TheOwner: TComponent; ATCPServer: TTCPServer;
  ASettings: TSettings);
begin
  inherited Create(TheOwner);
  FTCPServer := ATCPServer;
  FSettings := ASettings;
end;

procedure TfStart.FormShow(Sender: TObject);
begin
  LoadSettings;
end;

procedure TfStart.LoadSettings;
begin
  with FSettings do
  begin
    Load;
    edBindIP.Text := FSettings.BindConnection.FIP;
    spPort.Value := StrToIntDef(FSettings.BindConnection.FPort, 0);
    edPass.Text := FSettings.BindConnection.FPass;
    if FSettings.AutoStart then
    begin
      FAutoStartCnt := 5;
      bStart.Caption := 'Start (' + IntToStr(FAutoStartCnt) + ')';
      tmAutoStart.Enabled := True;
    end;
  end;
end;

procedure TfStart.SaveSettings;
begin
  with FSettings do
  begin
    FSettings.BindConnection.FIP := edBindIP.Text;
    FSettings.BindConnection.FPort := IntToStr(spPort.Value);
    FSettings.BindConnection.FPass := edPass.Text;
    Save;
  end;
end;

procedure TfStart.EnableDisable(AEnable: Boolean);
begin
  edBindIP.Enabled := AEnable;
  lbBindIP.Enabled := AEnable;
  spPort.Enabled := AEnable;
  lbPort.Enabled := AEnable;
  edPass.Enabled := AEnable;
  lbPass.Enabled := AEnable;
  bStart.Enabled := AEnable;
  bCancel.Enabled := AEnable;
end;

procedure TfStart.bStartClick(Sender: TObject);
var
  LogStr: String;
begin
  if Trim(edBindIP.Text) = '' then
  begin
    edBindIp.SetFocus;
    MessageDlgEx('Please enter a valid ip or host.', mtInformation, [mbOk], [], False, Self);
    Exit;
  end;
  if (tmAutoStart.Enabled) then
  begin
    tmAutoStart.Enabled := False;
    bStart.Caption := 'Start';
  end;

  EnableDisable(False);
  Self.Invalidate;
  Application.ProcessMessages;
  SaveSettings;
  FTCPServer.HideInternalMessages := FSettings.HideInternalMessages;
  FTCPServer.HideErrorMessages := FSettings.HideErrorMessages;
  FTCPServer.DownloadDirectory := FSettings.DownloadDirectory;
  FTCPServer.Start(FSettings.BindConnection);
  EnableDisable(True);
  if FTCPServer.Active then
    ModalResult := mrOk
  else
  begin
    LogStr := 'Cannot start server.';
    fMain.sb.Tag := 2;
    fMain.sb.Panels[0].Text := LogStr;
    MessageDlgEx(LogStr + ' Error message: ' + sLineBreak + '"' + FTCPServer.ErrMsg + '(' + IntToStr(FTCPServer.ErrCode) + ')"', mtInformation, [mbOk], [], False, Self);
  end;
end;

procedure TfStart.tmAutoStartTimer(Sender: TObject);
begin
  Dec(FAutoStartCnt);
  bStart.Caption := 'Start (' + IntToStr(FAutoStartCnt) + ')';
  if FAutoStartCnt = 0 then
  begin
    tmAutoStart.Enabled := False;
    bStart.Caption := 'Start';
    bStartClick(bStart);
  end;
end;

procedure TfStart.edBindIPEnter(Sender: TObject);
begin
  if (FSettings.AutoStart) and (tmAutoStart.Enabled) then
  begin
    tmAutoStart.Enabled := False;
    bStart.Caption := 'Start';
  end;
end;


end.

